#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # NeuralTreeNet et coût
# 
# La classe [NeuralTreeNet](:ref:`cl-NeuralTreeNet`) convertit un arbre de décision en réseau de neurones. Si la conversion n'est pas exacte mais elle permet d'obtenir un modèle différentiable et apprenable avec un algorithme d'optimisation à base de gradient. Ce notebook compare le temps d'éxécution entre un arbre et le réseau de neurones.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[3]:


get_ipython().run_line_magic('load_ext', 'mlprodict')


# ## Jeux de données
# 
# On construit un jeu de données aléatoire.

# In[4]:


import numpy

X = numpy.random.randn(10000, 10)
y = X.sum(axis=1) / X.shape[1]
X = X.astype(numpy.float64)
y = y.astype(numpy.float64)


# In[5]:


middle = X.shape[0] // 2
X_train, X_test = X[:middle], X[middle:]
y_train, y_test = y[:middle], y[middle:]


# ## Caler un arbre de décision

# In[6]:


from sklearn.tree import DecisionTreeRegressor

tree = DecisionTreeRegressor(max_depth=7)
tree.fit(X_train, y_train)
tree.score(X_train, y_train), tree.score(X_test, y_test)


# In[7]:


from sklearn.metrics import r2_score
r2_score(y_test, tree.predict(X_test))


# Covnersion de l'arbre en réseau de neurones

# In[8]:


from pandas import DataFrame
from mlstatpy.ml.neural_tree import NeuralTreeNet, NeuralTreeNetRegressor

xe = X_test.astype(numpy.float32)
expected = tree.predict(xe)

nn = NeuralTreeNetRegressor(NeuralTreeNet.create_from_tree(tree, arch='compact'))
got = nn.predict(xe)
me = numpy.abs(got - expected).mean()
mx = numpy.abs(got - expected).max()
DataFrame([{"average absolute error": me, "max absolute error": mx}]).T


# La conversion est loin d'être parfaite. La raison vient du fait que les fonctions de seuil sont approchées par des fonctions sigmoïdes. Il suffit d'une erreur minime pour que la décision prenne un chemin différent dans l'arbre et soit complètement différente.

# ## Conversion au format ONNX

# In[9]:


from mlprodict.onnx_conv import to_onnx

onx_tree = to_onnx(tree, X[:1].astype(numpy.float32))
onx_nn = to_onnx(nn, X[:1].astype(numpy.float32))


# Le réseau de neurones peut être représenté comme suit.

# In[10]:


get_ipython().run_line_magic('onnxview', 'onx_nn')


# In[11]:


from mlprodict.plotting.text_plot import onnx_simple_text_plot
print(onnx_simple_text_plot(onx_nn))


# ## Temps de calcul des prédictions

# In[12]:


from mlprodict.onnxrt import OnnxInference

oinf_tree = OnnxInference(onx_tree, runtime='onnxruntime1')
oinf_nn = OnnxInference(onx_nn, runtime='onnxruntime1')

get_ipython().run_line_magic('timeit', 'tree.predict(xe)')


# In[13]:


get_ipython().run_line_magic('timeit', "oinf_tree.run({'X': xe})")


# In[14]:


get_ipython().run_line_magic('timeit', "oinf_nn.run({'X': xe})")


# Le temps de calcul est nettement plus long pour le réseau de neurones. Si l'arbre de décision a une profondeur de *d*, l'arbre de décision va faire exactement *d* comparaisons. Le réseau de neurones quant à lui évalue tous les seuils pour chaque prédiction, soit $2^d$. Vérifions cela en faisant variable la profondeur.

# ## Temps de calcul en fonction de la profondeur

# In[15]:


from tqdm import tqdm
from cpyquickhelper.numbers import measure_time

data = []
for d in tqdm(range(2, 10)):
    tree = DecisionTreeRegressor(max_depth=d)
    tree.fit(X_train, y_train)
    obs = measure_time(lambda: tree.predict(xe), number=20, repeat=20)
    obs.update(dict(d=d, exp='skl'))
    data.append(obs)
    
    nn = NeuralTreeNetRegressor(NeuralTreeNet.create_from_tree(tree, arch='compact'))
    
    onx_tree = to_onnx(tree, X[:1].astype(numpy.float32))
    onx_nn = to_onnx(nn, X[:1].astype(numpy.float32))
    oinf_tree = OnnxInference(onx_tree, runtime='onnxruntime1')
    oinf_nn = OnnxInference(onx_nn, runtime='onnxruntime1')
    
    obs = measure_time(lambda: oinf_tree.run({'X': xe}), number=10, repeat=10)
    obs.update(dict(d=d, exp='onx_tree'))
    data.append(obs)

    obs = measure_time(lambda: oinf_nn.run({'X': xe}), number=10, repeat=10)
    obs.update(dict(d=d, exp='onx_nn'))
    data.append(obs)

df = DataFrame(data)
df


# In[16]:


piv = df.pivot('d', 'exp', 'average')
piv.plot(logy=True, title="Temps de calcul en fonction de la profondeur");


# L'hypothèse est vérifiée.

# In[17]: